@@ -35,12 +35,8 @@
 struct crypt_device {
 	char *type;
 
-	char *device;
-	char *metadata_device;
-
-	/* loopback automatic detach helpers */
-	int loop_device_fd;
-	int loop_metadata_device_fd;
+	struct device *device;
+	struct device *metadata_device;
 
 	struct volume_key *volume_key;
 	uint64_t timeout;
@@ -156,11 +152,22 @@ void logger(struct crypt_device *cd, int level, const char *file,
 	free(target);
 }
 
-static const char *mdata_device(struct crypt_device *cd)
+static const char *mdata_device_path(struct crypt_device *cd)
+{
+	return device_path(cd->metadata_device ?: cd->device);
+}
+
+/* internal only */
+struct device *crypt_metadata_device(struct crypt_device *cd)
 {
 	return cd->metadata_device ?: cd->device;
 }
 
+struct device *crypt_data_device(struct crypt_device *cd)
+{
+	return cd->device;
+}
+
 static int init_crypto(struct crypt_device *ctx)
 {
 	int r;
@@ -307,7 +314,7 @@ int PLAIN_activate(struct crypt_device *cd,
 		.uuid   = crypt_get_uuid(cd),
 		.size   = size,
 		.flags  = flags,
-		.data_device = crypt_get_device_name(cd),
+		.data_device = crypt_data_device(cd),
 		.u.crypt  = {
 			.cipher = NULL,
 			.vk     = vk,
@@ -321,8 +328,8 @@ int PLAIN_activate(struct crypt_device *cd,
 	else
 		device_check = DEV_EXCL;
 
-	r = device_check_and_adjust(cd, dmd.data_device, device_check,
-				    &dmd.size, &dmd.u.crypt.offset, &dmd.flags);
+	r = device_block_adjust(cd, dmd.data_device, device_check,
+				dmd.u.crypt.offset, &dmd.size, &dmd.flags);
 	if (r)
 		return r;
 
@@ -337,10 +344,10 @@ int PLAIN_activate(struct crypt_device *cd,
 	log_dbg("Trying to activate PLAIN device %s using cipher %s.",
 		name, dmd.u.crypt.cipher);
 
-	r = dm_create_device(name, CRYPT_PLAIN, &dmd, 0);
+	r = dm_create_device(cd, name, CRYPT_PLAIN, &dmd, 0);
 
 	// FIXME
-	if (!cd->plain_uuid && dm_query_device(name, DM_ACTIVE_UUID, &dmd) >= 0)
+	if (!cd->plain_uuid && dm_query_device(cd, name, DM_ACTIVE_UUID, &dmd) >= 0)
 		cd->plain_uuid = CONST_CAST(char*)dmd.uuid;
 
 	free(dm_cipher);
@@ -417,7 +424,7 @@ static int volume_key_by_terminal_passphrase(struct crypt_device *cd, int keyslo
 		if(r < 0)
 			goto out;
 
-		r = LUKS_open_key_with_hdr(mdata_device(cd), keyslot, passphrase_read,
+		r = LUKS_open_key_with_hdr(keyslot, passphrase_read,
 					   passphrase_size_read, &cd->hdr, vk, cd);
 		if (r == -EPERM)
 			eperm = 1;
@@ -504,44 +511,6 @@ const char *crypt_get_dir(void)
 	return dm_get_dir();
 }
 
-static int set_device_or_loop(const char *device_org, char **device, int *loop_fd)
-{
-	int r, readonly = 0;
-
-	if (!device_org)
-		return 0;
-
-	r = device_ready(NULL, device_org, O_RDONLY);
-	if (r == -ENOTBLK) {
-		*device = crypt_loop_get_device();
-		log_dbg("Not a block device, %s%s.", *device ?
-			"using free loop device " : "no free loop device found",
-			*device ?: "");
-		if (!*device) {
-			log_err(NULL, _("Cannot find a free loopback device.\n"));
-			return -ENOSYS;
-		}
-
-		/* Keep the loop open, dettached on last close. */
-		*loop_fd = crypt_loop_attach(*device, device_org, 0, 1, &readonly);
-		if (*loop_fd == -1) {
-			log_err(NULL, _("Attaching loopback device failed "
-				"(loop device with autoclear flag is required).\n"));
-			return -EINVAL;
-		}
-
-		r = device_ready(NULL, *device, O_RDONLY);
-	}
-
-	if (r < 0)
-		return -ENOTBLK;
-
-	if (!*device && device_org && !(*device = strdup(device_org)))
-		return -ENOMEM;
-
-	return 0;
-}
-
 int crypt_init(struct crypt_device **cd, const char *device)
 {
 	struct crypt_device *h = NULL;
@@ -556,17 +525,12 @@ int crypt_init(struct crypt_device **cd, const char *device)
 		return -ENOMEM;
 
 	memset(h, 0, sizeof(*h));
-	h->loop_device_fd = -1;
-	h->loop_metadata_device_fd = -1;
 
-	r = set_device_or_loop(device, &h->device, &h->loop_device_fd);
+	r = device_alloc(&h->device, device);
 	if (r < 0)
 		goto bad;
 
-	if (dm_init(h, 1) < 0) {
-		r = -ENOSYS;
-		goto bad;
-	}
+	dm_backend_init();
 
 	h->iteration_time = 1000;
 	h->password_verify = 0;
@@ -575,12 +539,7 @@ int crypt_init(struct crypt_device **cd, const char *device)
 	*cd = h;
 	return 0;
 bad:
-
-	if (h) {
-		if (h->loop_device_fd != -1)
-			close(h->loop_device_fd);
-		free(h->device);
-	}
+	device_free(h->device);
 	free(h);
 	return r;
 }
@@ -593,13 +552,13 @@ static int crypt_check_data_device_size(struct crypt_device *cd)
 	/* Check data device size, require at least one sector */
 	size_min = crypt_get_data_offset(cd) << SECTOR_SHIFT ?: SECTOR_SIZE;
 
-	r = device_size(crypt_get_device_name(cd), &size);
+	r = device_size(cd->device, &size);
 	if (r < 0)
 		return r;
 
 	if (size < size_min) {
 		log_err(cd, _("Header detected but device %s is too small.\n"),
-			crypt_get_device_name(cd));
+			device_path(cd->device));
 		return -EINVAL;
 	}
 
@@ -608,8 +567,8 @@ static int crypt_check_data_device_size(struct crypt_device *cd)
 
 int crypt_set_data_device(struct crypt_device *cd, const char *device)
 {
-	char *data_device = NULL;
-	int r, loop_fd = -1;
+	struct device *dev = NULL;
+	int r;
 
 	log_dbg("Setting ciphertext data device to %s.", device ?: "(none)");
 
@@ -622,21 +581,16 @@ int crypt_set_data_device(struct crypt_device *cd, const char *device)
 	if (!cd->device || !device)
 		return -EINVAL;
 
-	r = set_device_or_loop(device, &data_device, &loop_fd);
+	r = device_alloc(&dev, device);
 	if (r < 0)
 		return r;
 
 	if (!cd->metadata_device) {
 		cd->metadata_device = cd->device;
-		cd->loop_metadata_device_fd = cd->loop_device_fd;
-	} else {
-		free(cd->device);
-		if (cd->loop_device_fd != -1)
-			close(cd->loop_device_fd);
-	}
+	} else
+		device_free(cd->device);
 
-	cd->device = data_device;
-	cd->loop_device_fd = loop_fd;
+	cd->device = dev;
 
 	return crypt_check_data_device_size(cd);
 }
@@ -650,7 +604,7 @@ static int _crypt_load_luks1(struct crypt_device *cd, int require_header, int re
 	if (r < 0)
 		return r;
 
-	r = LUKS_read_phdr(mdata_device(cd), &hdr, require_header, repair, cd);
+	r = LUKS_read_phdr(&hdr, require_header, repair, cd);
 	if (r < 0)
 		return r;
 
@@ -671,14 +625,13 @@ static int _crypt_load_verity(struct crypt_device *cd, struct crypt_params_verit
 	if (r < 0)
 		return r;
 
-	if (params->flags & CRYPT_VERITY_NO_HEADER)
+	if (params && params->flags & CRYPT_VERITY_NO_HEADER)
 		return -EINVAL;
 
 	if (params)
 		sb_offset = params->hash_area_offset;
 
-	r = VERITY_read_sb(cd, mdata_device(cd), sb_offset,
-			   &cd->verity_uuid, &cd->verity_hdr);
+	r = VERITY_read_sb(cd, sb_offset, &cd->verity_uuid, &cd->verity_hdr);
 	if (r < 0)
 		return r;
 
@@ -706,10 +659,11 @@ static int _init_by_name_crypt(struct crypt_device *cd, const char *name)
 	char cipher[MAX_CIPHER_LEN], cipher_mode[MAX_CIPHER_LEN];
 	int key_nums, r;
 
-	r = dm_query_device(name, DM_ACTIVE_DEVICE |
-				   DM_ACTIVE_UUID |
-				   DM_ACTIVE_CRYPT_CIPHER |
-				   DM_ACTIVE_CRYPT_KEYSIZE, &dmd);
+	r = dm_query_device(cd, name,
+			DM_ACTIVE_DEVICE |
+			DM_ACTIVE_UUID |
+			DM_ACTIVE_CRYPT_CIPHER |
+			DM_ACTIVE_CRYPT_KEYSIZE, &dmd);
 	if (r < 0)
 		goto out;
 
@@ -740,7 +694,7 @@ static int _init_by_name_crypt(struct crypt_device *cd, const char *name)
 			cd->loopaes_key_size = dmd.u.crypt.vk->keylength / key_nums;
 		}
 	} else if (isLUKS(cd->type)) {
-		if (mdata_device(cd)) {
+		if (crypt_metadata_device(cd)) {
 			r = _crypt_load_luks1(cd, 0, 0);
 			if (r < 0) {
 				log_dbg("LUKS device header does not match active device.");
@@ -763,8 +717,8 @@ static int _init_by_name_crypt(struct crypt_device *cd, const char *name)
 	}
 out:
 	crypt_free_volume_key(dmd.u.crypt.vk);
+	device_free(dmd.data_device);
 	free(CONST_CAST(void*)dmd.u.crypt.cipher);
-	free(CONST_CAST(void*)dmd.data_device);
 	free(CONST_CAST(void*)dmd.uuid);
 	return r;
 }
@@ -778,10 +732,11 @@ static int _init_by_name_verity(struct crypt_device *cd, const char *name)
 	};
 	int r;
 
-	r = dm_query_device(name, DM_ACTIVE_DEVICE |
-				   DM_ACTIVE_UUID |
-				   DM_ACTIVE_VERITY_HASH_DEVICE |
-				   DM_ACTIVE_VERITY_PARAMS, &dmd);
+	r = dm_query_device(cd, name,
+				DM_ACTIVE_DEVICE |
+				DM_ACTIVE_UUID |
+				DM_ACTIVE_VERITY_HASH_DEVICE |
+				DM_ACTIVE_VERITY_PARAMS, &dmd);
 	if (r < 0)
 		goto out;
 
@@ -801,12 +756,10 @@ static int _init_by_name_verity(struct crypt_device *cd, const char *name)
 		cd->verity_hdr.flags = params.flags;
 		cd->verity_hdr.salt_size = params.salt_size;
 		cd->verity_hdr.salt = params.salt;
-		if (!(cd->metadata_device = strdup(dmd.u.verity.hash_device)))
-			r = -ENOMEM;
+		cd->metadata_device = dmd.u.verity.hash_device;
 	}
 out:
-	free(CONST_CAST(void*)dmd.u.verity.hash_device);
-	free(CONST_CAST(void*)dmd.data_device);
+	device_free(dmd.data_device);
 	free(CONST_CAST(void*)dmd.uuid);
 	return r;
 }
@@ -830,7 +783,7 @@ int crypt_init_by_name_and_header(struct crypt_device **cd,
 		return -ENODEV;
 	}
 
-	r = dm_query_device(name, DM_ACTIVE_DEVICE | DM_ACTIVE_UUID, &dmd);
+	r = dm_query_device(NULL, name, DM_ACTIVE_DEVICE | DM_ACTIVE_UUID, &dmd);
 	if (r < 0)
 		goto out;
 
@@ -839,7 +792,7 @@ int crypt_init_by_name_and_header(struct crypt_device **cd,
 	if (header_device) {
 		r = crypt_init(cd, header_device);
 	} else {
-		r = crypt_init(cd, dmd.data_device);
+		r = crypt_init(cd, device_path(dmd.data_device));
 
 		/* Underlying device disappeared but mapping still active */
 		if (!dmd.data_device || r == -ENOTBLK)
@@ -848,7 +801,7 @@ int crypt_init_by_name_and_header(struct crypt_device **cd,
 
 		/* Underlying device is not readable but crypt mapping exists */
 		if (r == -ENOTBLK) {
-			free(CONST_CAST(void*)dmd.data_device);
+			device_free(dmd.data_device);
 			dmd.data_device = NULL;
 			r = crypt_init(cd, NULL);
 		}
@@ -872,7 +825,7 @@ int crypt_init_by_name_and_header(struct crypt_device **cd,
 		log_dbg("Active device has no UUID set, some parameters are not set.");
 
 	if (header_device) {
-		r = crypt_set_data_device(*cd, dmd.data_device);
+		r = crypt_set_data_device(*cd, device_path(dmd.data_device));
 		if (r < 0)
 			goto out;
 	}
@@ -888,7 +841,7 @@ out:
 		crypt_free(*cd);
 		*cd = NULL;
 	}
-	free(CONST_CAST(void*)dmd.data_device);
+	device_free(dmd.data_device);
 	free(CONST_CAST(void*)dmd.uuid);
 	return r;
 }
@@ -954,7 +907,7 @@ static int _crypt_format_luks1(struct crypt_device *cd,
 	unsigned long required_alignment = DEFAULT_DISK_ALIGNMENT;
 	unsigned long alignment_offset = 0;
 
-	if (!mdata_device(cd)) {
+	if (!crypt_metadata_device(cd)) {
 		log_err(cd, _("Can't format LUKS without device.\n"));
 		return -EINVAL;
 	}
@@ -973,39 +926,50 @@ static int _crypt_format_luks1(struct crypt_device *cd,
 
 	if (params && params->data_device) {
 		cd->metadata_device = cd->device;
-		if (!(cd->device = strdup(params->data_device)))
+		cd->device = NULL;
+		if (device_alloc(&cd->device, params->data_device) < 0)
 			return -ENOMEM;
 		required_alignment = params->data_alignment * SECTOR_SIZE;
 	} else if (params && params->data_alignment) {
 		required_alignment = params->data_alignment * SECTOR_SIZE;
 	} else
-		get_topology_alignment(cd->device, &required_alignment,
+		device_topology_alignment(cd->device,
+				       &required_alignment,
 				       &alignment_offset, DEFAULT_DISK_ALIGNMENT);
 
+	/* Check early if we cannot allocate block device for key slot access */
+	r = device_block_adjust(cd, cd->device, DEV_OK, 0, NULL, NULL);
+	if(r < 0)
+		return r;
+
 	r = LUKS_generate_phdr(&cd->hdr, cd->volume_key, cipher, cipher_mode,
 			       (params && params->hash) ? params->hash : "sha1",
 			       uuid, LUKS_STRIPES,
 			       required_alignment / SECTOR_SIZE,
 			       alignment_offset / SECTOR_SIZE,
 			       cd->iteration_time, &cd->PBKDF2_per_sec,
-			       cd->metadata_device, cd);
+			       cd->metadata_device ? 1 : 0, cd);
 	if(r < 0)
 		return r;
 
 	/* Wipe first 8 sectors - fs magic numbers etc. */
-	r = crypt_wipe(mdata_device(cd), 0, 8 * SECTOR_SIZE, CRYPT_WIPE_ZERO, 1);
+	r = crypt_wipe(crypt_metadata_device(cd), 0, 8 * SECTOR_SIZE, CRYPT_WIPE_ZERO, 1);
 	if(r < 0) {
 		if (r == -EBUSY)
 			log_err(cd, _("Cannot format device %s which is still in use.\n"),
-				mdata_device(cd));
-		else
+				mdata_device_path(cd));
+		else if (r == -EACCES) {
+			log_err(cd, _("Cannot format device %s, permission denied.\n"),
+				mdata_device_path(cd));
+			r = -EINVAL;
+		} else
 			log_err(cd, _("Cannot wipe header on device %s.\n"),
-				mdata_device(cd));
+				mdata_device_path(cd));
 
 		return r;
 	}
 
-	r = LUKS_write_phdr(mdata_device(cd), &cd->hdr, cd);
+	r = LUKS_write_phdr(&cd->hdr, cd);
 
 	return r;
 }
@@ -1016,7 +980,7 @@ static int _crypt_format_loopaes(struct crypt_device *cd,
 				 size_t volume_key_size,
 				 struct crypt_params_loopaes *params)
 {
-	if (!mdata_device(cd)) {
+	if (!crypt_metadata_device(cd)) {
 		log_err(cd, _("Can't format LOOPAES without device.\n"));
 		return -EINVAL;
 	}
@@ -1052,7 +1016,7 @@ static int _crypt_format_verity(struct crypt_device *cd,
 	int r = 0, hash_size;
 	uint64_t data_device_size;
 
-	if (!mdata_device(cd)) {
+	if (!crypt_metadata_device(cd)) {
 		log_err(cd, _("Can't format VERITY without device.\n"));
 		return -EINVAL;
 	}
@@ -1104,7 +1068,8 @@ static int _crypt_format_verity(struct crypt_device *cd,
 		return -ENOMEM;
 
 	cd->verity_hdr.flags = params->flags;
-	cd->verity_hdr.hash_name = strdup(params->hash_name);
+	if (!(cd->verity_hdr.hash_name = strdup(params->hash_name)))
+		return -ENOMEM;
 	cd->verity_hdr.data_device = NULL;
 	cd->verity_hdr.data_block_size = params->data_block_size;
 	cd->verity_hdr.hash_block_size = params->hash_block_size;
@@ -1112,7 +1077,9 @@ static int _crypt_format_verity(struct crypt_device *cd,
 	cd->verity_hdr.hash_type = params->hash_type;
 	cd->verity_hdr.flags = params->flags;
 	cd->verity_hdr.salt_size = params->salt_size;
-	cd->verity_hdr.salt = malloc(params->salt_size);
+	if (!(cd->verity_hdr.salt = malloc(params->salt_size)))
+		return -ENOMEM;
+
 	if (params->salt)
 		memcpy(CONST_CAST(char*)cd->verity_hdr.salt, params->salt,
 		       params->salt_size);
@@ -1123,7 +1090,7 @@ static int _crypt_format_verity(struct crypt_device *cd,
 		return r;
 
 	if (params->flags & CRYPT_VERITY_CREATE_HASH) {
-		r = VERITY_create(cd, &cd->verity_hdr, cd->device, mdata_device(cd),
+		r = VERITY_create(cd, &cd->verity_hdr,
 				  cd->verity_root_hash, cd->verity_root_hash_size);
 		if (r)
 			return r;
@@ -1138,8 +1105,7 @@ static int _crypt_format_verity(struct crypt_device *cd,
 				return r;
 		}
 
-		r = VERITY_write_sb(cd, mdata_device(cd),
-				    cd->verity_hdr.hash_area_offset,
+		r = VERITY_write_sb(cd, cd->verity_hdr.hash_area_offset,
 				    cd->verity_uuid,
 				    &cd->verity_hdr);
 	}
@@ -1165,7 +1131,7 @@ int crypt_format(struct crypt_device *cd,
 		return -EINVAL;
 	}
 
-	log_dbg("Formatting device %s as type %s.", mdata_device(cd) ?: "(none)", type);
+	log_dbg("Formatting device %s as type %s.", mdata_device_path(cd) ?: "(none)", type);
 
 	r = init_crypto(cd);
 	if (r < 0)
@@ -1203,9 +1169,9 @@ int crypt_load(struct crypt_device *cd,
 	int r;
 
 	log_dbg("Trying to load %s crypt type from device %s.",
-		requested_type ?: "any", mdata_device(cd) ?: "(none)");
+		requested_type ?: "any", mdata_device_path(cd) ?: "(none)");
 
-	if (!mdata_device(cd))
+	if (!crypt_metadata_device(cd))
 		return -EINVAL;
 
 	if (!requested_type || isLUKS(requested_type)) {
@@ -1234,9 +1200,9 @@ int crypt_repair(struct crypt_device *cd,
 	int r;
 
 	log_dbg("Trying to repair %s crypt type from device %s.",
-		requested_type ?: "any", mdata_device(cd) ?: "(none)");
+		requested_type ?: "any", mdata_device_path(cd) ?: "(none)");
 
-	if (!mdata_device(cd))
+	if (!crypt_metadata_device(cd))
 		return -EINVAL;
 
 	if (requested_type && !isLUKS(requested_type))
@@ -1269,7 +1235,7 @@ int crypt_resize(struct crypt_device *cd, const char *name, uint64_t new_size)
 
 	log_dbg("Resizing device %s to %" PRIu64 " sectors.", name, new_size);
 
-	r = dm_query_device(name, DM_ACTIVE_DEVICE | DM_ACTIVE_CRYPT_CIPHER |
+	r = dm_query_device(cd, name, DM_ACTIVE_DEVICE | DM_ACTIVE_CRYPT_CIPHER |
 				  DM_ACTIVE_UUID | DM_ACTIVE_CRYPT_KEYSIZE |
 				  DM_ACTIVE_CRYPT_KEY, &dmd);
 	if (r < 0) {
@@ -1282,8 +1248,8 @@ int crypt_resize(struct crypt_device *cd, const char *name, uint64_t new_size)
 		goto out;
 	}
 
-	r = device_check_and_adjust(cd, dmd.data_device, DEV_OK, &new_size,
-				    &dmd.u.crypt.offset, &dmd.flags);
+	r = device_block_adjust(cd, dmd.data_device, DEV_OK,
+				dmd.u.crypt.offset, &new_size, &dmd.flags);
 	if (r)
 		goto out;
 
@@ -1293,7 +1259,7 @@ int crypt_resize(struct crypt_device *cd, const char *name, uint64_t new_size)
 		r = 0;
 	} else {
 		dmd.size = new_size;
-		r = dm_create_device(name, cd->type, &dmd, 1);
+		r = dm_create_device(cd, name, cd->type, &dmd, 1);
 	}
 out:
 	if (dmd.target == DM_CRYPT) {
@@ -1315,19 +1281,19 @@ int crypt_set_uuid(struct crypt_device *cd, const char *uuid)
 
 	if (uuid && !strncmp(uuid, cd->hdr.uuid, sizeof(cd->hdr.uuid))) {
 		log_dbg("UUID is the same as requested (%s) for device %s.",
-			uuid, mdata_device(cd));
+			uuid, mdata_device_path(cd));
 		return 0;
 	}
 
 	if (uuid)
-		log_dbg("Requested new UUID change to %s for %s.", uuid, mdata_device(cd));
+		log_dbg("Requested new UUID change to %s for %s.", uuid, mdata_device_path(cd));
 	else
-		log_dbg("Requested new UUID refresh for %s.", mdata_device(cd));
+		log_dbg("Requested new UUID refresh for %s.", mdata_device_path(cd));
 
 	if (!crypt_confirm(cd, _("Do you really want to change UUID of device?")))
 		return -EPERM;
 
-	return LUKS_hdr_uuid_set(mdata_device(cd), &cd->hdr, uuid, cd);
+	return LUKS_hdr_uuid_set(&cd->hdr, uuid, cd);
 }
 
 int crypt_header_backup(struct crypt_device *cd,
@@ -1344,9 +1310,9 @@ int crypt_header_backup(struct crypt_device *cd,
 		return r;
 
 	log_dbg("Requested header backup of device %s (%s) to "
-		"file %s.", mdata_device(cd), requested_type, backup_file);
+		"file %s.", mdata_device_path(cd), requested_type, backup_file);
 
-	return LUKS_hdr_backup(backup_file, mdata_device(cd), &cd->hdr, cd);
+	return LUKS_hdr_backup(backup_file, &cd->hdr, cd);
 }
 
 int crypt_header_restore(struct crypt_device *cd,
@@ -1363,26 +1329,21 @@ int crypt_header_restore(struct crypt_device *cd,
 		return r;
 
 	log_dbg("Requested header restore to device %s (%s) from "
-		"file %s.", mdata_device(cd), requested_type, backup_file);
+		"file %s.", mdata_device_path(cd), requested_type, backup_file);
 
-	return LUKS_hdr_restore(backup_file, mdata_device(cd), &cd->hdr, cd);
+	return LUKS_hdr_restore(backup_file, &cd->hdr, cd);
 }
 
 void crypt_free(struct crypt_device *cd)
 {
 	if (cd) {
-		log_dbg("Releasing crypt device %s context.", mdata_device(cd));
+		log_dbg("Releasing crypt device %s context.", mdata_device_path(cd));
 
-		if (cd->loop_device_fd != -1)
-			close(cd->loop_device_fd);
-		if (cd->loop_metadata_device_fd != -1)
-			close(cd->loop_metadata_device_fd);
-
-		dm_exit();
+		dm_backend_exit();
 		crypt_free_volume_key(cd->volume_key);
 
-		free(cd->device);
-		free(cd->metadata_device);
+		device_free(cd->device);
+		device_free(cd->metadata_device);
 		free(cd->type);
 
 		/* used in plain device only */
@@ -1414,7 +1375,7 @@ int crypt_suspend(struct crypt_device *cd,
 
 	log_dbg("Suspending volume %s.", name);
 
-	if (!isLUKS(cd->type)) {
+	if (!cd || !isLUKS(cd->type)) {
 		log_err(cd, _("This operation is supported only for LUKS device.\n"));
 		r = -EINVAL;
 		goto out;
@@ -1426,10 +1387,9 @@ int crypt_suspend(struct crypt_device *cd,
 		return -EINVAL;
 	}
 
-	if (!cd && dm_init(NULL, 1) < 0)
-		return -ENOSYS;
+	dm_backend_init();
 
-	r = dm_status_suspended(name);
+	r = dm_status_suspended(cd, name);
 	if (r < 0)
 		goto out;
 
@@ -1439,14 +1399,13 @@ int crypt_suspend(struct crypt_device *cd,
 		goto out;
 	}
 
-	r = dm_suspend_and_wipe_key(name);
+	r = dm_suspend_and_wipe_key(cd, name);
 	if (r == -ENOTSUP)
 		log_err(cd, "Suspend is not supported for device %s.\n", name);
 	else if (r)
 		log_err(cd, "Error during suspending device %s.\n", name);
 out:
-	if (!cd)
-		dm_exit();
+	dm_backend_exit();
 	return r;
 }
 
@@ -1467,7 +1426,7 @@ int crypt_resume_by_passphrase(struct crypt_device *cd,
 		goto out;
 	}
 
-	r = dm_status_suspended(name);
+	r = dm_status_suspended(cd, name);
 	if (r < 0)
 		return r;
 
@@ -1477,14 +1436,14 @@ int crypt_resume_by_passphrase(struct crypt_device *cd,
 	}
 
 	if (passphrase) {
-		r = LUKS_open_key_with_hdr(mdata_device(cd), keyslot, passphrase,
-					   passphrase_size, &cd->hdr, &vk, cd);
+		r = LUKS_open_key_with_hdr(keyslot, passphrase, passphrase_size,
+					   &cd->hdr, &vk, cd);
 	} else
 		r = volume_key_by_terminal_passphrase(cd, keyslot, &vk);
 
 	if (r >= 0) {
 		keyslot = r;
-		r = dm_resume_and_reinstate_key(name, vk->keylength, vk->key);
+		r = dm_resume_and_reinstate_key(cd, name, vk->keylength, vk->key);
 		if (r == -ENOTSUP)
 			log_err(cd, "Resume is not supported for device %s.\n", name);
 		else if (r)
@@ -1516,7 +1475,7 @@ int crypt_resume_by_keyfile_offset(struct crypt_device *cd,
 		goto out;
 	}
 
-	r = dm_status_suspended(name);
+	r = dm_status_suspended(cd, name);
 	if (r < 0)
 		return r;
 
@@ -1534,13 +1493,13 @@ int crypt_resume_by_keyfile_offset(struct crypt_device *cd,
 	if (r < 0)
 		goto out;
 
-	r = LUKS_open_key_with_hdr(mdata_device(cd), keyslot, passphrase_read,
+	r = LUKS_open_key_with_hdr(keyslot, passphrase_read,
 				   passphrase_size_read, &cd->hdr, &vk, cd);
 	if (r < 0)
 		goto out;
 
 	keyslot = r;
-	r = dm_resume_and_reinstate_key(name, vk->keylength, vk->key);
+	r = dm_resume_and_reinstate_key(cd, name, vk->keylength, vk->key);
 	if (r)
 		log_err(cd, "Error during resuming device %s.\n", name);
 out:
@@ -1596,7 +1555,7 @@ int crypt_keyslot_add_by_passphrase(struct crypt_device *cd,
 		}
 	} else if (passphrase) {
 		/* Passphrase provided, use it to unlock existing keyslot */
-		r = LUKS_open_key_with_hdr(mdata_device(cd), CRYPT_ANY_SLOT, passphrase,
+		r = LUKS_open_key_with_hdr(CRYPT_ANY_SLOT, passphrase,
 					   passphrase_size, &cd->hdr, &vk, cd);
 	} else {
 		/* Passphrase not provided, ask first and use it to unlock existing keyslot */
@@ -1605,7 +1564,7 @@ int crypt_keyslot_add_by_passphrase(struct crypt_device *cd,
 		if (r < 0)
 			goto out;
 
-		r = LUKS_open_key_with_hdr(mdata_device(cd), CRYPT_ANY_SLOT, password,
+		r = LUKS_open_key_with_hdr(CRYPT_ANY_SLOT, password,
 					   passwordLen, &cd->hdr, &vk, cd);
 		crypt_safe_free(password);
 	}
@@ -1623,7 +1582,7 @@ int crypt_keyslot_add_by_passphrase(struct crypt_device *cd,
 			goto out;
 	}
 
-	r = LUKS_set_key(mdata_device(cd), keyslot, new_password, new_passwordLen,
+	r = LUKS_set_key(keyslot, new_password, new_passwordLen,
 			 &cd->hdr, vk, cd->iteration_time, &cd->PBKDF2_per_sec, cd);
 	if(r < 0) goto out;
 
@@ -1682,7 +1641,7 @@ int crypt_keyslot_add_by_keyfile_offset(struct crypt_device *cd,
 		if (r < 0)
 			goto out;
 
-		r = LUKS_open_key_with_hdr(mdata_device(cd), CRYPT_ANY_SLOT, password, passwordLen,
+		r = LUKS_open_key_with_hdr(CRYPT_ANY_SLOT, password, passwordLen,
 					   &cd->hdr, &vk, cd);
 	}
 
@@ -1699,7 +1658,7 @@ int crypt_keyslot_add_by_keyfile_offset(struct crypt_device *cd,
 	if (r < 0)
 		goto out;
 
-	r = LUKS_set_key(mdata_device(cd), keyslot, new_password, new_passwordLen,
+	r = LUKS_set_key(keyslot, new_password, new_passwordLen,
 			 &cd->hdr, vk, cd->iteration_time, &cd->PBKDF2_per_sec, cd);
 out:
 	crypt_safe_free(password);
@@ -1765,7 +1724,7 @@ int crypt_keyslot_add_by_volume_key(struct crypt_device *cd,
 		passphrase_size = new_passwordLen;
 	}
 
-	r = LUKS_set_key(mdata_device(cd), keyslot, passphrase, passphrase_size,
+	r = LUKS_set_key(keyslot, passphrase, passphrase_size,
 			 &cd->hdr, vk, cd->iteration_time, &cd->PBKDF2_per_sec, cd);
 out:
 	crypt_safe_free(new_password);
@@ -1795,7 +1754,7 @@ int crypt_keyslot_destroy(struct crypt_device *cd, int keyslot)
 		return -EINVAL;
 	}
 
-	return LUKS_del_key(mdata_device(cd), keyslot, &cd->hdr, cd);
+	return LUKS_del_key(keyslot, &cd->hdr, cd);
 }
 
 // activation/deactivation of device mapping
@@ -1851,7 +1810,7 @@ int crypt_activate_by_passphrase(struct crypt_device *cd,
 	} else if (isLUKS(cd->type)) {
 		/* provided passphrase, do not retry */
 		if (passphrase) {
-			r = LUKS_open_key_with_hdr(mdata_device(cd), keyslot, passphrase,
+			r = LUKS_open_key_with_hdr(keyslot, passphrase,
 						   passphrase_size, &cd->hdr, &vk, cd);
 		} else
 			r = volume_key_by_terminal_passphrase(cd, keyslot, &vk);
@@ -1923,7 +1882,7 @@ int crypt_activate_by_keyfile_offset(struct crypt_device *cd,
 			  &passphrase_size_read, keyfile, keyfile_offset, keyfile_size);
 		if (r < 0)
 			goto out;
-		r = LUKS_open_key_with_hdr(mdata_device(cd), keyslot, passphrase_read,
+		r = LUKS_open_key_with_hdr(keyslot, passphrase_read,
 					   passphrase_size_read, &cd->hdr, &vk, cd);
 		if (r < 0)
 			goto out;
@@ -2033,8 +1992,7 @@ int crypt_activate_by_volume_key(struct crypt_device *cd,
 			return -EINVAL;
 		}
 
-		r = VERITY_activate(cd, name, mdata_device(cd),
-				    volume_key, volume_key_size,
+		r = VERITY_activate(cd, name, volume_key, volume_key_size,
 				    &cd->verity_hdr, CRYPT_ACTIVATE_READONLY);
 
 		if (r == -EPERM) {
@@ -2064,13 +2022,13 @@ int crypt_deactivate(struct crypt_device *cd, const char *name)
 
 	log_dbg("Deactivating volume %s.", name);
 
-	if (!cd && dm_init(NULL, 1) < 0)
-		return -ENOSYS;
+	if (!cd)
+		dm_backend_init();
 
 	switch (crypt_status(cd, name)) {
 		case CRYPT_ACTIVE:
 		case CRYPT_BUSY:
-			r = dm_remove_device(name, 0, 0);
+			r = dm_remove_device(cd, name, 0, 0);
 			break;
 		case CRYPT_INACTIVE:
 			log_err(cd, _("Device %s is not active.\n"), name);
@@ -2082,7 +2040,7 @@ int crypt_deactivate(struct crypt_device *cd, const char *name)
 	}
 
 	if (!cd)
-		dm_exit();
+		dm_backend_exit();
 
 	return r;
 }
@@ -2115,7 +2073,7 @@ int crypt_volume_key_get(struct crypt_device *cd,
 		if (r < 0)
 			log_err(cd, _("Cannot retrieve volume key for plain device.\n"));
 	} else if (isLUKS(cd->type)) {
-		r = LUKS_open_key_with_hdr(mdata_device(cd), keyslot, passphrase,
+		r = LUKS_open_key_with_hdr(keyslot, passphrase,
 					passphrase_size, &cd->hdr, &vk, cd);
 
 	} else
@@ -2212,13 +2170,13 @@ crypt_status_info crypt_status(struct crypt_device *cd, const char *name)
 {
 	int r;
 
-	if (!cd && dm_init(NULL, 1) < 0)
-		return CRYPT_INVALID;
+	if (!cd)
+		dm_backend_init();
 
-	r = dm_status_device(name);
+	r = dm_status_device(cd, name);
 
 	if (!cd)
-		dm_exit();
+		dm_backend_exit();
 
 	if (r < 0 && r != -ENODEV)
 		return CRYPT_INVALID;
@@ -2243,7 +2201,7 @@ static int _luks_dump(struct crypt_device *cd)
 {
 	int i;
 
-	log_std(cd, "LUKS header information for %s\n\n", mdata_device(cd));
+	log_std(cd, "LUKS header information for %s\n\n", mdata_device_path(cd));
 	log_std(cd, "Version:       \t%d\n", cd->hdr.version);
 	log_std(cd, "Cipher name:   \t%s\n", cd->hdr.cipherName);
 	log_std(cd, "Cipher mode:   \t%s\n", cd->hdr.cipherMode);
@@ -2286,7 +2244,7 @@ static int _luks_dump(struct crypt_device *cd)
 
 static int _verity_dump(struct crypt_device *cd)
 {
-	log_std(cd, "VERITY header information for %s\n", mdata_device(cd));
+	log_std(cd, "VERITY header information for %s\n", mdata_device_path(cd));
 	log_std(cd, "UUID:            \t%s\n", cd->verity_uuid ?: "");
 	log_std(cd, "Hash type:       \t%u\n", cd->verity_hdr.hash_type);
 	log_std(cd, "Data blocks:     \t%" PRIu64 "\n", cd->verity_hdr.data_size);
@@ -2365,7 +2323,12 @@ const char *crypt_get_uuid(struct crypt_device *cd)
 
 const char *crypt_get_device_name(struct crypt_device *cd)
 {
-	return cd->device;
+	const char *path = device_block_path(cd->device);
+
+	if (!path)
+		path = device_path(cd->device);
+
+	return path;
 }
 
 int crypt_get_volume_key_size(struct crypt_device *cd)
@@ -2431,6 +2394,17 @@ int crypt_keyslot_max(const char *type)
 	return -EINVAL;
 }
 
+int crypt_keyslot_area(struct crypt_device *cd,
+	int keyslot,
+	uint64_t *offset,
+	uint64_t *length)
+{
+	if (!isLUKS(cd->type))
+		return -EINVAL;
+
+	return LUKS_keyslot_area(&cd->hdr, keyslot, offset, length);
+}
+
 const char *crypt_get_type(struct crypt_device *cd)
 {
 	return cd->type;
@@ -2442,8 +2416,8 @@ int crypt_get_verity_info(struct crypt_device *cd,
 	if (!isVERITY(cd->type) || !vp)
 		return -EINVAL;
 
-	vp->data_device = cd->device;
-	vp->hash_device = mdata_device(cd);
+	vp->data_device = device_path(cd->device);
+	vp->hash_device = mdata_device_path(cd);
 	vp->hash_name = cd->verity_hdr.hash_name;
 	vp->salt = cd->verity_hdr.salt;
 	vp->salt_size = cd->verity_hdr.salt_size;
@@ -2463,7 +2437,7 @@ int crypt_get_active_device(struct crypt_device *cd __attribute__((unused)),
 	struct crypt_dm_active_device dmd;
 	int r;
 
-	r = dm_query_device(name, 0, &dmd);
+	r = dm_query_device(cd, name, 0, &dmd);
 	if (r < 0)
 		return r;
 
